@@ -20,7 +20,7 @@ with 'Net::HTTP::Knork::Role::Middleware';
 has 'client' => ( is => 'lazy', );
 
 # option that allows one to pass optional parameters that are not specified
-# in the spore 'optional_params' section for a given method
+# in the spore 'optional_params' or 'optional_payload' section for a given method.
 
 has 'lax_optionals' => ( is => 'rw', default => sub {0} );
 
@@ -90,6 +90,16 @@ has 'spore_rx' => (
     }
 );
 
+has 'encoding' => ( 
+    is => 'rw',
+    default => sub { undef }, 
+); 
+
+has 'decoding' => ( 
+    is => 'rw',
+    default => sub { undef },
+);
+
 
 # Change the namespace of a given instance, so that there won't be any
 # method collision between two instances
@@ -180,21 +190,7 @@ sub make_sub_from_spec {
         }
         my %method_args = %{$meth_spec};
         my $method      = $method_args{method};
-        my $payload =
-          ( defined $param_spec{spore_payload} )
-          ? delete $param_spec{spore_payload}
-          : delete $param_spec{payload};
 
-        if ( $method_args{required_payload} && !$payload ) {
-            croak "this method requires a payload and no payload is provided";
-        }
-        if ( $payload
-            && ( $method !~ /^(?:POST|PUT|PATCH)$/i ) )
-        {
-            croak "payload requires a PUT, PATCH or POST method";
-        }
-
-        $payload //= undef;
 
         if ( $method_args{required_params} ) {
             foreach my $required ( @{ $method_args{required_params} } ) {
@@ -204,8 +200,20 @@ sub make_sub_from_spec {
                 }
             }
         }
+        if ( $method_args{required_payload} || $method_args{optional_payload}) {
+            unless ($method =~ /POST|PUT|PATCH/i) { 
+                croak 'Payloads are to be used with POST/PUT/PATCH HTTP methods';
+            }
+            foreach my $required ( @{ $method_args{required_payload} } ) {
+                if ( !grep { $required eq $_ } keys %param_spec ) {
+                    croak
+                      "Parameter '$required' is required for the payload but is missing";
+                }
+            }
+        }
 
         my $params;
+        my $payload; 
         foreach ( @{ $method_args{required_params} } ) {
             push @$params, $_, delete $param_spec{$_};
         }
@@ -214,10 +222,39 @@ sub make_sub_from_spec {
             push @$params, $_, delete $param_spec{$_}
               if ( defined( $param_spec{$_} ) );
         }
+        foreach ( @{ $method_args{required_payload}} ) { 
+            $payload->{$_} = delete $param_spec{$_};
+        }
+        foreach (@{ $method_args{optional_payload} } ) {
+            $payload->{$_} = delete $param_spec{$_}
+              if ( defined( $param_spec{$_} ) );
+        }
+        
         if (%param_spec) {
             if ( $self->lax_optionals ) {
-                foreach ( keys %param_spec ) {
-                    push @$params, $_, delete $param_spec{$_};
+                my $option = $self->lax_optionals;
+                if ( $option eq 'params' ) {
+                    foreach ( keys %param_spec ) {
+                        push @$params, $_, delete $param_spec{$_};
+                    }
+                }
+                else {
+                    if ( $option eq 'payload' ) {
+                        if ( $method =~ /POST|PUT|PATCH/i ) {
+                            foreach ( keys %param_spec ) {
+                                $payload->{$_} = delete $param_spec{$_};
+                            }
+                        }
+                        else {
+                            carp "Discarding the remaining parameters: cannot"
+                              . " use lax payload with a method not supporting payloads";
+                        }
+                    }
+                    else {
+                        carp
+                          "Discarding the remaining parameters since 'lax_optionals'"
+                          . " is set, but not to 'params' or 'payload'";
+                    }
                 }
             }
         }
@@ -246,12 +283,17 @@ sub make_sub_from_spec {
             'spore.errors'     => *STDERR,
             'spore.url_scheme' => $base_url->scheme,
             'spore.userinfo'   => $base_url->userinfo,
-
+            'spore.encode'     => $self->encoding, 
+            'spore.decode'     => $self->decoding,
         };
         $self->env($env);
         my $request      = $self->request->finalize();
         my $raw_response = $self->perform_request($request);
-        return $self->generate_response($raw_response);
+        my $knork_response = $self->request->new_response(
+            $raw_response->code, $raw_response->message, $raw_response->headers,
+            $raw_response->content
+        );
+        return $self->generate_response($knork_response);
     };
 }
 
@@ -266,15 +308,11 @@ sub generate_response {
     my $self           = shift;
     my $raw_response   = shift;
     my $orig_response  = shift;
-    my $knork_response = $self->request->new_response(
-        $raw_response->code, $raw_response->message, $raw_response->headers,
-        $raw_response->content
-    );
     if ( defined($orig_response) ) {
-        $knork_response->raw_body( $orig_response->content )
-          unless defined( ( $knork_response->raw_body ) );
+        $raw_response->raw_body( $orig_response->content )
+          unless defined( ( $raw_response->raw_body ) );
     }
-    return $knork_response;
+    return $raw_response;
 }
 
 
@@ -282,15 +320,18 @@ sub generate_response {
 1;
 
 __END__
+
 =pod
 
+=encoding UTF-8
+
 =head1 NAME
 
 Net::HTTP::Knork - Lightweight implementation of Spore specification
 
 =head1 VERSION
 
-version 0.11
+version 0.20
 
 =head1 SYNOPSIS
 
@@ -315,28 +356,44 @@ version 0.11
     my $knork = Net::HTTP::Knork->new(spec => $spec);
 
     # make a GET to 'http://example.com/test/bar'
-    my $resp = $knork->test({ foo => bar});
+    my $resp = $knork->test({ foo => 'bar'});
 
 =head1 DESCRIPTION
 
-Net::HTTP::Knork is a module that aims to be compatible with L<the Spore specification|https://github.com/SPORE/specifications/blob/master/spore_description.pod>.
+Net::HTTP::Knork is a module that was inspired by <the Spore specification|https://github.com/SPORE/specifications/blob/master/spore_description.pod>.
 So it is like L<Net::HTTP::Spore> but with some differences.
 
-=head2 Moo !
+=head2 JSON or Perl hash specifications
 
-When I was working with Net::HTTP::Spore, I found it hard to get around all the magic done with Moose.
-So this implementation aims at having something more lightweight.
+Specifications can be written either in a JSON file, string, or as a Perl hash (no YAML support).
+On top of that, every specification passed is validated against a Spore specification embedded, using L<Data::Rx>.
 
-=head2 Specifications
+=head2 No middleware modules
 
-Specifications can be written either in a JSON file, string, or as a pure Perl hash.
-On top of that, every specification passed is validated against the Spore specification, using L<Data::Rx>.
-
-=head2 No middlewares
-
-This module does not provide middlewares as in L<Net::HTTP::Spore>, but there are some ways around it that should fit basic needs.
+This module does not provide middlewares as in L<Net::HTTP::Spore>, but you can alter the default behaviour on requests/responses.
 See Middlewares below
 
+=head2 Behaviour of requests with payloads (POST/PUT/PATCH)
+
+The original behaviour of Net::HTTP::Spore with POST/PATCH/PUT was not clear enough, so I made some modifications: 
+
+=over
+
+=item * 
+I changed the meaning of required_payload and optional_payload: now they contain the fields required or optional in the payload
+
+=item * 
+I removed altogether the support for form-data and payload
+
+=item * 
+Consequently I removed the { payload => ... } that was passed to methods when making a request with a payload 
+
+=item * 
+The default encoding on payload requests is now set to 'application/x-www-form-urlencoded'. I don't plan on support 'multipart/form-data'
+right now, but if needed, it can be added by using 'encoding' and 'decoding' in the constructor.
+
+=back
+
 =head2 HTTP::Response compliant
 
 All the responses returned by Knork are objects from a class extending L<HTTP::Response>.
@@ -347,6 +404,11 @@ This means that you can basically manipulate any response returned by a Knork cl
 No assumptions are made regarding the responses you may receive from an API.
 It means that, contrary to L<Net::HTTP::Spore>, the code won't just die if the API returns a 4XX. This also implies that you should always check the responses returned.
 
+=head2 Moo !
+
+When I was working with Net::HTTP::Spore, I found it hard to get around all the magic done with Moose.
+So this implementation tries to be lightweight.
+
 =head1 METHODS
 
 =over
@@ -370,12 +432,56 @@ Other constructor options:
 default_params: hash specifying default parameters to pass on every requests.
 
     # pass foo=bar on every request
-    my $client = Net::HTTP::Knork->new(spec => 'some_file.json', default_params => {foo => bar});
+    my $client = Net::HTTP::Knork->new(spec => 'some_file.json', default_params => {foo => 'bar'});
 
 =item *
 
 client: L<LWP::UserAgent> HTTP client. Automatically created if not passed.
 
+=item *
+
+lax_optionals: a string parameter that can either be set to 'payload' or 'params'. When this parameter
+is set, the remaining parameters passed to any method will be either put in the URL or in the payload.
+The default behaviour is to ignore any parameter that is not explicitly put in the spec. 
+
+=item * 
+encoding/decoding: those parameters allow you to apply an encoding/decoding on the requests/responses. 
+Encoding or decoding is applied BEFORE the middlewares and can be changed on the fly (attributes are rw).
+
+=over
+
+=item * 
+encoding: a subref that will be applied before the 'middlewares' to encode a request in a certain way. Takes a 
+Net::HTTP::Knork::Request object as an argument and MUST return a Net::HTTP::Knork::Request object. 
+When using the encoding attribute, the content is set to a Perl hash containing the keys/values needed for 
+the POST request.
+
+    use JSON::MaybeXS;
+    my $client = Net::HTTP::Knork->new($spec => 'some_spec.json', 
+        encoding => sub {  
+            my $req = shift; 
+            my $content = $req->content; 
+            $req->content(encode_json ($content));
+            return $req;
+        }
+    );
+
+=item *
+decoding: a subref that will be applied before the 'middlewares' to decode a request in a certain way. Takes a 
+Net::HTTP::Knork::Response object as an argument and MUST return a Net::HTTP::Knork object.
+
+    use JSON::MaybeXS;
+    my $client = Net::HTTP::Knork->new($spec => 'some_spec.json', 
+        decoding => sub {  
+            my $req = shift; 
+            my $content = $req->content; 
+            $req->content(decode_json ($content));
+            return $req;
+        }
+    );
+
+=back
+
 =back
 
 =item make_sub_from_spec
@@ -397,30 +503,26 @@ You might want to do that if you want to create new subs with parameters you can
 
     use strict;
     use warnings;
-    use JSON::MaybeXS;
     use Net::HTTP::Knork;
-    # create a client with a middleware that encode requests to json and
-    # decode responses from json
     my $client = Net::HTTP::Knork->new(spec => '/path/to/spec.json');
     $client->add_middleware(
         {   on_request => sub {
                 my $self = shift;
                 my $req = shift;
-                $req->header( 'Content-Type' => 'application/json' );
-                $req->header( 'Accept'       => 'application/json' );
-                $req->content( encode_json( $req->content ) );
+                # alter the request in some way
                 return $req;
             },
             on_response => sub {
                 my $self = shift;
                 my $resp = shift;
-                $resp->content(decode_json( $resp->content ) );
+                # alter the response in some way
                 return $resp;
               }
         }
     );
 
-Although middlewares cannot be created as in L<Net::HTTP::Spore>, there is still the possibility to declare subs that will be executed either on requests or responses.
+Although middlewares cannot be created as in L<Net::HTTP::Spore>, there is still the possibility to declare subs that will 
+be executed either before the request is sent or before the response is processed. 
 To accomplish this, it installs modifiers around some core functions in L<Net::HTTP::Knork>, using L<Class::Method::Modifiers>.
 
 =head2 Limitations
@@ -531,4 +633,3 @@ Fabrice Gabolde - <fabrice.gabolde@gmail.com>
 =back
 
 =cut
-
